#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['minpiler']

package_data = \
{'': ['*']}

extras_require = \
{":python_version <= '3.6'": ['dataclasses']}

entry_points = \
{'console_scripts': ['minpiler = minpiler.cmdline:run']}

setup(name='minpiler',
      version='0.3.2',
      description='Compile python code to Mindustry microprocessor instructions',
      author='Vitalik Verhovodov',
      author_email='knifeslaughter@gmail.com',
      url='https://github.com/neumond/minpiler',
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6',
     )
